(require (file "../spur.scm"))
(require (file "../phrase.scm"))

;;;  This is a variant of Opening to Glassworks by Philip Glass.
;;;  It shows how multiple metros can be used for polyrhythm.
;;;  It also shows how to schedule changes to the velocities of notes.

(define stop? #f)                                 ; stop playing?
(define bpm 92)                                   ; beats per minute
(define *bar* (/ (* 4 *minute*) bpm))             ; time per bar   
(define start (+ (now) (* 1 *second*)))           ; when to start playing

;;  velocity releated settings
(define gvelocity -5)                             ; global velocity adjustment
(define drange 10)                                ; dynamic range (+ ve)
(define rrange 10)                                ; random range (+ ve)

;; adjust the velocity of a note in a phrase using gvelocity + randomize
(define (mod-vel n) (set-note-vel n + (+ gvelocity (rand (- rrange) rrange))))

;; schedule all the phrases in LPHRS in sequence using the metronome MET
(define (pphrase met . lphrs)
  (let loop ((time start) (phrs (apply append lphrs)))
    (unless (or (null? phrs) stop?)
      (let ((dur (met 'dur (/ (phrase-length (car phrs)) 96))))
        (schedule-phrase time 1 (map mod-vel (car phrs)) (met 'get-tempo))
        (schedule (+ time (* dur .8)) loop (+ time dur) (cdr phrs)) ))))

;; (repl N1 item1 N2 item2 ...) generate phrases from N1 copies of item1 +
;; N2 copies of item2 etc.  The items are strings or other lists of phrases
(define (repl . args)
  (if (null? args) '()
      (append (if (string? (cadr args)) 
                  (make-list (car args) (string->phrase (cadr args)))
                  (apply append (make-list (car args) (cadr args))))
              (apply repl (cddr args)))))

;; generate phrases from N-1 copies of item1+item2 and 1 copy of item1+item3
(define (dupl n item1 item2 item3)
  (define seqfn (if (string? item1) string-append append))
  (append (repl (- n 1) (seqfn item1 item2)) (repl 1 (seqfn item1 item3))))

;; (crescend B1 V1 B2 V2 ...) add drange*Vi to velocities of notes from bar Bi
;; This is done smoothly by adding the V amount drange times within bar B
(define (crescend . items)
  (define pre (- start (* 100 *milli*)))    ; just before the notes
  (define div (/ *bar* drange))             ; fraction of a bar
  (let loop ((items items))
    (define (update) (set! gvelocity (+ gvelocity (cadr items))))
    (unless (null? items)
      (let iloop ((i drange) (time (+ pre (* *bar* (- (car items) 1)))))
        (if (= i 0) (loop (cddr items))
           (begin 
             (schedule time update)
             (schedule (+ time (* .8 div)) iloop (- i 1) (+ time div))))))))

;;;  -------   The music starts here                                ; BARS

;; phrases for the long notes of the left hand
(define (long)
  (pphrase (make-metro (/ bpm 4)) 
    (dupl 4 "F3mp:192,Eb:96," "Db3" "Bb2")                          ; 1--16
    (repl 4 "C3mp,Bb2,Bb,G")                                        ; 17--32
    (dupl 3 "F3:192,Ab:96," "Bb3" "Bb3:48,Ab") (repl 1 "G3:192,F")  ; 33-48
    (repl 1 "F3mf")))

;; phrases for the other notes of the left hand
(define (left)
  (pphrase (make-metro (* bpm 2)) 
    (repl 4 (repl 8 "Ab3,C4" 8 "G3,Bb"))                            ; 1-16 
    (repl 4 (repl 8 "F3,Ab" 8 "Eb3,G"))                             ; 17-32
    (dupl 4 (repl 8 "Bb3,D4") (repl 8 "C4,Eb") (repl 8 "Bb3,D4"))   ; 33-48
    (repl 1 "D4mf")))

;; phrases for the right hand
(define (right)
  (pphrase (make-metro (* bpm 3)) 
     (repl 4 (repl 6 "F4,Ab" 6 "C5,F4" 6 "G4,C5" 6 "Ab4,C5"))       ; 1-16
     (dupl 4 (repl 6 "C4,F" 6 "D4,F" 6 "Bb3,Eb4")
             (repl 6 "C4,Eb") (repl 6 "Bb3,Eb4"))                   ; 17-32 
     (repl 3 (repl 6 "D5mf,F4m" 6 "Ab4,D5" 12 "Eb5mf,Ab4m"))
     (repl 1 (repl 12 "D5mf,G4m" 6 "D5,Ab4" 6 "D5,Bb4mp"))          ; 33-48
     (repl 1 "Bb4mf")))

;; global changes in the dynamics of the notes played
(define (dynamics)
  (crescend  
      2 2  4 -2   6 3  8 -3  10 1 12 -2  14 3 16 -4                 ; 1-16
     17 2 20 -2  21 3 24 -3  25 1 28 -2  29 3 32 -1                 ; 17-32
     34 2 36 -1  38 2 40 -2  42 1 44 -2  46 1 48 -2))               ; 33-48

;;;  -------   The main function:  start the four parts + the pedal

(define (main . strs) 
  (random-seed (if (null? strs) 4741114 (string->number (car strs))))
  (with-midi-schedule
    (schedule start control-change 1 64 127)                   ; pedal down
    (schedule (+ start (* 50 *bar*)) control-change 1 64 0)    ; pedal up
    (left) (right) (long) (dynamics) ))
