#lang racket

;;;  This defines an event scheduler to be used with the coremidi library.
;;;  If the timing is critical, racket should be run sans garbage collection.

(require (lib "defmacro.ss") "midiout.scm")
(provide (all-from-out "midiout.scm") 
     now *milli* *second* *minute* *hour* real->integer rand make-metro
     initialize-schedule execute-schedule 
     map-schedule adjust-schedule quantize-schedule
     open-midi-device close-midi-device
     schedule callback schedule* schedule-note
     with-midi-out with-midi-schedule with-schedule define*)

;; work in units of milliseconds
(define now current-inexact-milliseconds)
(define *milli* 1)
(define *second* (* *milli* 1000))
(define *minute* (* *second* 60))
(define *hour* (* *minute* 60))
(define default-end-wait (* 30 *milli*))          ; wait before stopping

;; Two useful Impromptu functions
(define (real->integer x) (inexact->exact (round x)))
(define (rand . args)                ; this is called "random" in Impromptu 
  (if (null? args) (random)
      (if (null? (cdr args)) (list-ref (car args) (random (length (car args))))
          (+ (car args) (random (- (cadr args) (car args)))))))

;; this is a variant of make-metro by Andrew Sorensen that returns a
;; function that responds to various requests involving beats
(define (make-metro bpm)
  (let ((ms (/ *minute* bpm)) (offset (now)))
    (lambda (sym . args)
      (let ((val (if (null? args) 1 (car args))))
        (cond 
         ((number? sym) (+ offset (* sym ms)))      ; beats -> time
         ((eq? sym 'dur) (* val ms))                ; beats -> duration
         ((eq? sym 'get-tempo) (/ *minute* ms))     ; return bpm
         ((eq? sym 'get-beat)                       ; next beat mod val
          (* val (ceiling (/ (- (now) offset) (* ms val)))))
         ((eq? sym 'set-tempo)                      ; reset bpm
          (let ((new (/ *minute* val)))
            (set! offset (/ (+ (* (now) (- ms new)) (* offset new)) ms))
            (set! ms new)))
         (else (error sym "wrong argument for metro")) )))))

;; an event consists of a time, a thunk, and a next event (or #f)
(define-macro (make-event time thunk next) `(mcons (cons ,time ,thunk) ,next))
(define-macro (ev-time ev) `(car (mcar ,ev)))
(define-macro (ev-thunk ev) `(cdr (mcar ,ev)))
(define-macro (ev-next ev) `(mcdr ,ev))
(define-macro (set-ev-pair! ev val) `(set-mcar! ,ev ,val))
(define-macro (set-ev-next! ev val) `(set-mcdr! ,ev ,val))

;; two pointers into the schedule
(define head-event #f)        ; the front of the schedule
(define tail-event #f)        ; somewhere near the rear

;; assigned during the schedule execution
(define next-event-time #f)   ; the time of the first event
(define waiting? #f)          ; paused in the executor?
(define pause-chan #f)        ; channel for releasing pauses

;; initialize the two event pointers and the execution variables
(define (initialize-schedule) 
  (collect-garbage)
  (set! head-event #f) (set! tail-event #f)
  (set! waiting? #f) (set! pause-chan (make-channel)))

;; sleep until next-event-time or until awakened by schedule*
(define (pause) 
  (set! waiting? #t)
  (sync (alarm-evt next-event-time) pause-chan)
  (set! waiting? #f))

;; execute each event in the schedule at its proper time
(define (execute-schedule [end-wait default-end-wait])
  (collect-garbage)
  (unless pause-chan (error "Schedule was not initialized"))
  (let loop ((time (now)))
    (if head-event                     ; any events in the schedule?
        (begin
          (set! next-event-time (ev-time head-event))
          (if (> next-event-time time) ; execute next event later?
              (pause)                  ; wait
              (let ((thunk (ev-thunk head-event)))
                (set! head-event (ev-next head-event))
                (thread thunk)))       ; do it now in a new thread
          (loop (now)))
        (begin                         ; last chance before stopping
          (set! next-event-time (+ time end-wait))
          (pause)                      ; allow for outstanding events
          (if head-event (loop (now)) (set! pause-chan #f))))))

;; create an event, insert it into its proper place in the schedule,
;; and wake a paused executor if the new event goes first
(define (schedule* time thunk)
  ; first adjust tail-event when head-event has moved beyond it
  (when (or (not head-event)
            (and tail-event (> (ev-time head-event) (ev-time tail-event))))
        (set! tail-event head-event))
  (let ((new (make-event time thunk #f)))
    ; a loop to search for the correct position for the new event
    (define (search-loop ev1 ev2)     ; ev1 < ev2  and  ev1 < time
      (if (and ev2 (> time (ev-time ev2))) 
          (search-loop ev2 (ev-next ev2))
          (begin (set! tail-event ev1)
                 (set-ev-next! new ev2)        
                 (set-ev-next! ev1 new))))
    (if (and tail-event (> time (ev-time tail-event)))
        (search-loop tail-event (ev-next tail-event))
        (if (and head-event (> time (ev-time head-event)))
            (search-loop head-event (ev-next head-event))
            (begin (set-ev-next! new head-event) 
                   (set! head-event new)))))
  (when (and waiting? (< time next-event-time))  ; if new event is now first 
        (channel-put pause-chan #t)))            ; release paused executor

(define-macro (schedule time fn . args) 
  `(schedule* ,time (lambda () (,fn ,@args))))

(define-macro (callback time fn . args)
  `(schedule* ,time (lambda () (,fn ,@args))))

;; map schedule to a list of (time,thunk) pairs, applying functions to each
(define (map-schedule time-fn thunk-fn)
  (define rfn (or time-fn (lambda (x) x)))
  (define tfn (or thunk-fn (lambda (x) x)))
  (let loop ((ev head-event))
    (if (not ev) '()
        (cons (cons (rfn (ev-time ev)) (tfn (ev-thunk ev)))
              (loop (ev-next ev))))))

;; apply functions to every time and thunk in the schedule
;; optional argument is lower bound on time
(define (adjust-schedule time-fn thunk-fn . args)
  (define low (if (null? args) 0 (car args)))
  (let loop ((ev head-event))
    (when (and ev (>= (ev-time ev) low))
      (set-ev-pair! ev (cons (time-fn (ev-time ev)) (thunk-fn (ev-thunk ev))))
      (loop (ev-next ev)))))

;; arrange the time of every event in the schedule to be a multiple of q
(define (quantize-schedule q . args)
  (when head-event
    (let ((start (+ (ev-time head-event) (if (null? args) 0 (car args)))))
      (adjust-schedule
       (lambda (time) 
         (let* ((dur (- time start))
                (div (/ dur q)))
           (+ start dur (* q (- (round div) div)))))
       (lambda (th) th)))))

;;;-------------------------------------------------------------------------
;;; MIDI

;; schedule a single note to play at a specific time
(define (schedule-note time chan pitch vel dur)
  (schedule time note-on chan pitch (real->integer vel))
  (schedule (+ time dur) note-off chan pitch))

(define-macro (with-midi-out . exprs)
  `(begin (open-midi-device) ,@exprs 
          (for ([i 16]) (control-change (+ i 1) 123 0)) ; all notes off
          (close-midi-device)))

(define-macro (with-schedule . exprs)
  `(begin (initialize-schedule) ,@exprs (execute-schedule)))

;; normal execution of a schedule
(define-macro (with-midi-schedule . exprs) 
  `(with-midi-out (with-schedule ,@exprs)))

;; for convenience: a new version of define that allows an empty definition
(define-macro (define* what . rest)
  (if (null? rest) 
      (if (symbol? what) `(define (,what . args) #f)
      `(define ,what #f))
      `(define ,what ,@rest)))
